"""
Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE

Contains large portions of:
https://github.com/Rapptz/discord.py/blob/master/discord/utils.py
The appropriate license is in LICENSE
"""

from __future__ import annotations

import abc
import asyncio
import base64
import collections
import functools
import html
import itertools
import re
import struct
import sys
import textwrap
import threading
import weakref
from collections.abc import (
    AsyncGenerator,
    AsyncIterable,
    Awaitable,
    Callable,
    Generator,
    Iterable,
    Iterator,
    Mapping,
    Sized,
)
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from inspect import getmembers, isawaitable
from io import BytesIO
from itertools import zip_longest
from operator import attrgetter
from types import MemberDescriptorType
from typing import (
    TYPE_CHECKING,
    Any,
    Concatenate,
    Final,
    Generic,
    Literal,
    Never,
    ParamSpec,
    TypeAlias,
    TypedDict,
    TypeVar,
    cast,
    overload,
)

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding, rsa
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

from ._const import JSON_LOADS, MISSING, UNIX_EPOCH, URL
from .enums import Type, _is_descriptor, classproperty as classproperty
from .id import _URL_START, ID, id64_from_url as id64_from_url, parse_id64 as parse_id64

if TYPE_CHECKING:
    from typing_extensions import Self, deprecated

    from .types.http import Coro, StrOrURL
    from .types.user import IndividualID


_T = TypeVar("_T")
_P = ParamSpec("_P")


def unpad(s: bytes, /) -> bytes:
    return s[: -s[-1]]


def symmetric_decrypt(text: bytes, key: bytes) -> bytes:
    cipher = Cipher(algorithms.AES(key), modes.ECB())
    decryptor = cipher.decryptor()
    iv = decryptor.update(text[:16])
    decryptor.finalize()

    cipher = Cipher(algorithms.AES(key), modes.CBC(iv))
    decryptor = cipher.decryptor()
    return unpad(decryptor.update(text[16:]))


STEAM_PUBLIC_KEY: Final = rsa.RSAPublicNumbers(
    17,
    int(
        (
            "1572435756163492767473017547683098671778311221560259237468446760604065883521072242173339019599191749864"
            "5577395742561473053175122897795413393419038630648254894306773660858554891146738442477393264257606729213"
            "7056263003121836768211312089498275802694267916711103128551999842076575732754013467986241640244933837449"
        )
    ),
)


def verify_signature(data: bytes, signature: bytes) -> bool:
    try:
        STEAM_PUBLIC_KEY.public_key().verify(signature, data, padding.PKCS1v15(), hashes.SHA1())
    except InvalidSignature:
        return False
    else:
        return True


@dataclass(slots=True)
class TradeURLInfo:
    id: IndividualID
    token: str | None = None

    @property
    def url(self) -> str:
        """The full trade URL."""
        url = URL.COMMUNITY / "tradeoffer/new" % {"partner": self.id.id}
        return str(url % {"token": self.token}) if self.token else str(url)

    def __str__(self) -> str:
        return self.url


def parse_trade_url(url: StrOrURL, /) -> TradeURLInfo | None:
    """Parses a trade URL for useful information.

    Parameters
    -----------
    url
        The trade URL to search.

    Returns
    -------
    .. source:: TradeURLInfo
    """
    if (
        match := re.match(
            (
                rf"{_URL_START}steamcommunity\.com/tradeoffer/new/?\?partner=(?P<user_id>\d{{,10}})"
                r"(?:&token=(?P<token>[\w-]{7,}))?"
            ),
            html.unescape(str(url)),
        )
    ) is None:
        return None

    return TradeURLInfo(ID(match["user_id"], type=Type.Individual), match["token"] or None)


_SelfT = TypeVar("_SelfT", covariant=True)
_T_co = TypeVar("_T_co", covariant=True)


class cached_property(Generic[_SelfT, _T_co]):
    __slots__ = ("function", "__doc__")

    def __init__(self, function: Callable[[_SelfT], _T_co], /):
        self.function = function
        self.__doc__: str | None = getattr(function, "__doc__", None)

    @overload
    def __get__(self, instance: None, _) -> Self: ...

    @overload
    def __get__(self, instance: _SelfT, _) -> _T_co:  # type: ignore
        ...

    def __get__(self, instance: _SelfT | None, _) -> _T_co | Self:
        if instance is None:
            return self

        value = self.function(instance)
        setattr(instance, self.function.__name__, value)

        return value

    if TYPE_CHECKING:

        def __set__(self, instance: _SelfT, value: _T_co) -> None:  # type: ignore
            ...


@overload
def cached_slot_property(function: Callable[[_SelfT], _T_co], /) -> CachedSlotProperty[_SelfT, _T_co]: ...


@overload
def cached_slot_property(name: str, /) -> Callable[[Callable[[_SelfT], _T_co]], CachedSlotProperty[_SelfT, _T_co]]: ...


def cached_slot_property(function_or_name: Any, /) -> Any:  # required to be a function to appease type checkers
    if isinstance(function_or_name, str):

        def wrapper(function: Callable[[_SelfT], _T_co]) -> CachedSlotProperty[_SelfT, _T_co]:
            return CachedSlotProperty(function, function_or_name)

        return wrapper
    return CachedSlotProperty(function_or_name, f"_{function_or_name.__name__}_cs")


class CachedSlotProperty(Generic[_SelfT, _T_co]):
    __slots__ = ("function", "__doc__", "name")

    def __init__(self, function: Callable[[_SelfT], _T_co], name: str):
        self.function = function
        self.__doc__: str | None = getattr(function, "__doc__", None)
        self.name = name
        if __debug__:  # sanity check slots are present
            try:
                slots = sys._getframe(2).f_locals["__slots__"]
            except KeyError:
                slots = sys._getframe(3).f_locals["__slots__"]
            assert name in slots, f"Slot {name} must exist"

    @overload
    def __get__(self, instance: None, _) -> Self: ...

    @overload
    def __get__(self, instance: _SelfT, _) -> _T_co:  # type: ignore
        ...

    def __get__(self, instance: _SelfT | None, _) -> _T_co | Self:
        if instance is None:
            return self

        try:
            return getattr(instance, self.name)
        except AttributeError:
            value = self.function(instance)
            setattr(instance, self.name, value)
            return value

    def __set__(self, instance: _SelfT, value: _T_co) -> None:  # type: ignore
        setattr(instance, self.name, value)


F_no_wait = TypeVar("F_no_wait", bound="Callable[Concatenate[Any, ...], Awaitable[None]]")
F_wait = TypeVar("F_wait", bound="Callable[Concatenate[Any, ...], Awaitable[Any]]")


@overload
def call_once(func: F_no_wait, /) -> F_no_wait: ...


@overload
def call_once(*, wait: Literal[False] = False) -> Callable[[F_no_wait], F_no_wait]: ...


@overload
def call_once(*, wait: Literal[True]) -> Callable[[F_wait], F_wait]: ...


def call_once(func: F_wait | None = None, /, *, wait: bool = False) -> F_wait | Callable[[F_wait], F_wait]:
    def get_inner(func: F_wait) -> F_wait:
        locks = weakref.WeakKeyDictionary[object, asyncio.Lock]()
        being_called = weakref.WeakSet[object]()

        if wait:
            futures = dict[object, asyncio.Future[Any]]()

            async def inner(self: Any, *args: Any, **kwargs: Any) -> None:
                try:
                    lock = locks[self]
                except KeyError:
                    lock = locks[self] = asyncio.Lock()

                async with lock:
                    if self in being_called:
                        return await futures[self]

                    being_called.add(self)
                    futures[self] = future = asyncio.get_running_loop().create_future()

                try:
                    res = await func(self, *args, **kwargs)
                except Exception as exc:
                    future.set_exception(exc)
                    raise exc
                else:
                    future.set_result(res)
                    return res
                finally:
                    being_called.remove(self)
                    del futures[self]

        else:

            async def inner(self: Any, *args: Any, **kwargs: Any) -> None:
                try:
                    lock = locks[self]
                except KeyError:
                    lock = locks[self] = asyncio.Lock()

                async with lock:
                    if self in being_called:
                        return

                    being_called.add(self)

                try:
                    await func(self, *args, **kwargs)
                finally:
                    being_called.remove(self)

        return cast(F_wait, functools.wraps(func)(inner))

    return get_inner if func is None else get_inner(func)


async def ainput(prompt: object = "", /) -> str:
    loop = asyncio.get_running_loop()
    future = loop.create_future()
    threading.Thread(target=lambda: loop.call_soon_threadsafe(future.set_result, input(prompt)), daemon=True).start()
    return await future


def _int_chunks(len: int, size: int) -> Generator[tuple[int, int], None, None]:
    idxs = range(0, len, size)
    second = iter(idxs)
    next(second)
    yield from zip_longest(idxs, second, fillvalue=len)


def update_class(
    instance: Any,
    new_instance: Any,
) -> Any:  # technically _T1 & _T2
    cls = instance.__class__
    is_descriptor = _is_descriptor

    for name, attr in getmembers(instance):
        attr = getattr(instance, name)
        cls_attr = getattr(cls, name, None)
        if (
            not is_descriptor(cls_attr)
            or isinstance(cls_attr, MemberDescriptorType)  # might be a slot member
            or (isinstance(cls_attr, property) and cls_attr.fset is not None)
        ) and not (name.startswith("__") and name.endswith("__")):
            try:
                setattr(new_instance, name, attr)
            except (AttributeError, TypeError):
                pass

    return new_instance


class DateTime:
    """UTC aware datetime factory functions."""

    @staticmethod
    def now() -> datetime:
        return datetime.now(timezone.utc)

    @staticmethod
    def from_timestamp(timestamp: float, /) -> datetime:
        # work around python/cpython#75395
        return UNIX_EPOCH + timedelta(seconds=timestamp)

    @staticmethod
    def strptime(input: str, format: str) -> datetime:
        dt = datetime.strptime(input, format)
        if dt.tzinfo is not None:
            return dt
        return dt.replace(tzinfo=timezone.utc)

    @staticmethod  # TODO actually make this reliable for languages other than english
    def parse_steam_date(input: str, /, *, full_month: bool = True) -> datetime | None:
        if ", " not in input:
            input += f"{input}, {DateTime.now().year}"  # assume current year

        british = input.split()[0].isdigit()  # starts with a number
        format = f"%d %{'B' if full_month else 'b'}, %Y" if british else f"%{'B' if full_month else 'b'} %d, %Y"
        try:
            return DateTime.strptime(input, format)
        except ValueError:
            return None


class AsyncInit(metaclass=abc.ABCMeta):
    __slots__ = ()

    @abc.abstractmethod
    async def __ainit__(self) -> None: ...

    async def __await_inner__(self) -> Self:
        await self.__ainit__()
        return self

    def __await__(self) -> Generator[Any, None, Self]:
        return self.__await_inner__().__await__()


@overload
async def race() -> None: ...
@overload
async def race(*coros: Awaitable[_T]) -> _T: ...
async def race(*coros: Awaitable[_T]) -> _T | None:
    class _Done(Exception):
        pass

    result = None

    async def run(coro: Awaitable[_T]) -> Never:
        nonlocal result
        result = await coro
        raise _Done

    try:
        async with asyncio.TaskGroup() as tg:
            for coro in coros:
                tg.create_task(run(coro))
    except* _Done:
        pass

    return result


PACK_FORMATS: Final = cast(Mapping[str, str], {
    "i8": "b",
    "u8": "B",
    "i16": "h",
    "u16": "H",
    "i32": "i",
    "u32": "I",
    "i64": "q",
    "u64": "Q",
    "long": "l",
    "ulong": "L",
    "f32": "f",
    "f64": "d",
})  # fmt: skip


class StructIOMeta(type):
    def __new__(mcs, name: str, bases: tuple[type, ...], namespace: dict[str, Any]) -> StructIOMeta:
        for method_name, format in PACK_FORMATS.items():
            exec(f"def write_{method_name}(self, item, /): self.write_struct('<{format}', item)", {}, namespace)
            exec(
                textwrap.dedent(
                    f"""def read_{method_name}(self):
                        (value,) = self.read_struct('<{format}', {struct.calcsize(f'<{format}')})
                        return value
                    """
                ),
                {},
                namespace,
            )

        return super().__new__(mcs, name, bases, namespace)


class StructIO(
    BytesIO, metaclass=type if TYPE_CHECKING else StructIOMeta
):  # type[BytesIO] is a subclass of ABCMeta at type time
    __slots__ = ()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(buffer={self.buffer!r}, position={self.position})"

    @property
    def buffer(self) -> bytes:
        return self.getvalue()

    @property
    def position(self) -> int:
        return self.tell()

    @position.setter
    def position(self, value: int) -> None:
        self.seek(value)

    def __len__(self) -> int:
        return len(self.getbuffer())

    def read_struct(self, format: str, /, position: int | None = None) -> tuple[Any, ...]:
        buffer = self.read(position or struct.calcsize(format))
        return struct.unpack(format, buffer)

    @overload
    def write_struct(self, format: str, to_write_1: Any, /, *to_write: Any) -> None:  # type: ignore
        ...

    def write_struct(self, format: str, /, *to_write: Any) -> None:
        self.write(struct.pack(format, *to_write))

    def read_cstring(self, terminator: bytes = b"\x00") -> bytes:
        data = self.getbuffer()[self.position :]
        for i, chars in enumerate(as_chunks(data, len(terminator))):
            if bytes(chars) == terminator:
                result = bytes(data[: i * len(terminator)])
                self.position += len(result) + len(terminator)
                return result

        raise RuntimeError("Reached end of buffer")

    def write_cstring(self, string: bytes, terminator: bytes = b"\x00") -> None:
        self.write(string)
        self.write(terminator)

    if TYPE_CHECKING:
        # added by the metaclass
        # fmt: off
        def read_i8(self) -> int: ...
        def write_i8(self, item: int, /) -> None: ...
        def read_u8(self) -> int: ...
        def write_u8(self, item: int, /) -> None: ...
        def read_i16(self) -> int: ...
        def write_i16(self, item: int, /) -> None: ...
        def read_u16(self) -> int: ...
        def write_u16(self, item: int, /) -> None: ...
        def read_i32(self) -> int: ...
        def write_i32(self, item: int, /) -> None: ...
        def read_u32(self) -> int: ...
        def write_u32(self, item: int, /) -> None: ...
        def read_i64(self) -> int: ...
        def write_i64(self, item: int, /) -> None: ...
        def read_u64(self) -> int: ...
        def write_u64(self, item: int, /) -> None: ...
        def read_f32(self) -> float: ...
        def write_f32(self, item: float, /) -> None: ...
        def read_f64(self) -> float: ...
        def write_f64(self, item: float, /) -> None: ...
        def read_long(self) -> int: ...
        def write_long(self, item: int, /) -> None: ...
        def read_ulong(self) -> int: ...
        def write_ulong(self, item: int, /) -> None: ...
        # fmt: on


_KT = TypeVar("_KT")
_VT = TypeVar("_VT")


class ChainMap(collections.ChainMap[_KT, _VT]):
    # this is different to the standard library's ChainMap because it is always O(n),
    # keys should be unique between maps
    def __delitem__(self, key: _KT) -> None:
        for map in self.maps:
            try:
                del map[key]
                return
            except KeyError:
                pass
        raise KeyError(key)

    def popitem(self) -> tuple[_KT, _VT]:
        for map in self.maps:
            try:
                return map.popitem()
            except KeyError:
                pass
        raise KeyError()

    @overload
    def pop(self, key: _KT) -> _VT: ...

    @overload
    def pop(self, key: _KT, default: _VT | _T) -> _VT | _T: ...

    def pop(self, key: _KT, default: _VT | _T = MISSING) -> _VT | _T:
        for map in self.maps:
            try:
                return map.pop(key)
            except KeyError:
                pass
        if default is not MISSING:
            return default
        raise KeyError(key)

    def clear(self) -> None:
        for map in self.maps:
            map.clear()


class JWTToken(TypedDict):
    iss: Literal["steam"]
    sub: str  # SteamID
    aud: list[str]
    exp: int
    nbf: int
    iat: int
    jti: str
    oat: int
    per: int
    ip_subject: str
    ip_confirmer: str


def decode_jwt(token: str, /) -> JWTToken:
    try:
        _, jwt, _ = token.split(".")
    except TypeError:
        raise ValueError("Invalid JWT") from None

    # python doesn't like the lack of padding on the end of the JWT so we need to add it back
    return JSON_LOADS(base64.b64decode(f"{jwt}==", altchars=b"-_"))


@dataclass(slots=True)
class BBCodeTag:
    name: str
    position: tuple[int, int]
    attributes: dict[str, str]
    inner: str


class BBCodeStr(str):
    tags: list[BBCodeTag]

    def __new__(cls, string: str, tags: list[BBCodeTag]) -> Self:
        self = super().__new__(cls, string)
        self.tags = tags
        return self


BB_CODE_RE: Final = re.compile(
    r"""
    (?<!\\)
    \[(?P<name>[\w]+)\s*
    (?P<attributes>(?:\s*\w*=[^]]+)*)
    \]
    (?P<inner>.*?)
    (?<!\\)
    \[/(?P=name)\]
    """,
    re.VERBOSE,
)

BB_CODE_ATTRIBUTES_RE: Final = re.compile(
    r"""
    (?P<key>[\w]*)
    \s*=\s*
    (?P<quote>["']?)
    (?P<value>[^ ]*)
    (?P=quote)
    """,
    re.VERBOSE,
)


def parse_bb_code(string: str, /) -> BBCodeStr:
    tags = list[BBCodeTag]()
    for match in BB_CODE_RE.finditer(string):
        tag = BBCodeTag(
            match["name"],
            match.span(),
            {key: value for key, _, value in BB_CODE_ATTRIBUTES_RE.findall(match["attributes"])},
            match["inner"],
        )
        tags.append(tag)

        new_start, new_end = tag.position
        tags += [  # AFAICT steam chat only ever has a maximum of one level of nesting on "inner" so this is fine
            BBCodeTag(
                match["name"],
                match.span(),
                {key: value for key, _, value in BB_CODE_ATTRIBUTES_RE.findall(match["attributes"])},
                match["inner"],
            )
            for match in BB_CODE_RE.finditer(string, new_start + 1, new_end - 1)
        ]

    return BBCodeStr(string, tags=tags)


CHAT_COMMANDS = "|".join(
    rf"{command_name}\s*"
    for command_name in (
        "me",
        "code",
        "pre",
        "giphy",
        "spoiler",
        "quote",
        "random",
        "flip",
        "sticker",
        "store",
    )
)
CHAT_COMMANDS_RE: Final = re.compile(rf"""/(?:{CHAT_COMMANDS})""")


def contains_chat_command(string: str, /) -> bool:
    return bool(CHAT_COMMANDS_RE.match(string))


if TYPE_CHECKING:
    todo = deprecated("This method is not yet implemented")
else:

    def todo(func):
        def wrapper(*args: Any, **kwargs: Any) -> None:
            raise NotImplementedError(f"{func.__name__} is not yet implemented")

        return wrapper


# everything below here is directly from discord.py's utils
# https://github.com/Rapptz/discord.py/blob/master/discord/utils.py
_Iter: TypeAlias = Iterable[_T] | AsyncIterable[_T]


if sys.version_info >= (3, 12):
    _chunk = itertools.batched
else:

    def _chunk(
        iterable: Iterable[_T],
        max_size: int,
        iter: Callable[[Iterable[_T]], Iterator[_T]] = iter,
        tuple: type[tuple[_T, ...]] = tuple,
        islice: type[itertools.islice[_T]] = itertools.islice,
        /,
    ) -> Generator[tuple[_T, ...], None, None]:
        it = iter(iterable)
        while batch := tuple(islice(it, max_size)):
            yield batch


async def _achunk(
    iterable: AsyncIterable[_T], max_size: int, len: Callable[[Sized], int] = len, /
) -> AsyncGenerator[tuple[_T, ...], None]:
    ret: list[_T] = []
    async for item in iterable:
        ret.append(item)
        if len(ret) == max_size:
            yield tuple(ret)
            ret = []
    if ret:
        yield tuple(ret)


@overload
def as_chunks(iterable: AsyncIterable[_T], /, max_size: int) -> AsyncGenerator[tuple[_T, ...], None]: ...


@overload
def as_chunks(iterable: Iterable[_T], /, max_size: int) -> Generator[tuple[_T, ...], None, None]: ...


def as_chunks(iterable: _Iter[_T], /, max_size: int) -> _Iter[tuple[_T, ...]]:
    """A helper function that collects an iterable into chunks of a given size.

    Parameters
    ----------
    iterable
        The iterable to chunk, can be sync or async.
    max_size
        The maximum chunk size.

    Warning
    -------
    The last chunk collected may not be as large as ``max_size``.

    Returns
    --------
    A new iterator which yields chunks of a given size.
    """
    if max_size <= 0:
        raise ValueError("max_size must be greater than 0")

    return _achunk(iterable, max_size) if hasattr(iterable, "__aiter__") else _chunk(iterable, max_size)  # type: ignore


@overload
def find(predicate: Callable[[_T], bool], iterable: AsyncIterable[_T], /) -> Coro[_T | None]: ...


@overload
def find(predicate: Callable[[_T], bool], iterable: Iterable[_T], /) -> _T | None: ...


def find(predicate: Callable[[_T], bool], iterable: _Iter[_T], /) -> _T | Coro[_T | None] | None:
    """A helper to return the first element found in the sequence.

    Examples
    --------
    .. code:: python

        first_active_offer = steam.utils.find(
            lambda trade: trade.state == TradeOfferState.Active,
            client.trades,
        )
        # how to get an object using a conditional

    Parameters
    -----------
    predicate
        A function that returns a boolean and takes an element from the ``iterable`` as its first argument.
    iterable
        The iterable to search through.

    Returns
    -------
    The first element from the ``iterable`` for which the ``predicate`` returns ``True`` or if no matching element was
    found returns ``None``.
    """

    if hasattr(iterable, "__aiter__"):  # isinstance(iterable, collections.abc.AsyncIterable) is too slow
        return anext((element async for element in iterable if predicate(element)), None)  # type: ignore
    else:
        return next((element for element in iterable if predicate(element)), None)  # type: ignore


def _get(
    iterable: Iterable[_T],
    all: Callable[[Iterable[bool]], bool] = all,
    attrgetter: type[attrgetter[_T]] = attrgetter,
    /,
    **attrs: Any,
) -> _T | None:
    # Special case the single element call
    if len(attrs) == 1:
        k, v = attrs.popitem()
        pred = attrgetter(k.replace("__", "."))
        return next((elem for elem in iterable if pred(elem) == v), None)

    converted = [(attrgetter(attr.replace("__", ".")), value) for attr, value in attrs.items()]

    return next((elem for elem in iterable if all(pred(elem) == value for pred, value in converted)), None)


def _aget(
    iterable: AsyncIterable[_T],
    all: Callable[[Iterable[bool]], bool] = all,
    attrgetter: type[attrgetter[_T]] = attrgetter,
    /,
    **attrs: Any,
) -> Coro[_T | None]:
    # Special case the single element call
    if len(attrs) == 1:
        k, v = attrs.popitem()
        pred = attrgetter(k.replace("__", "."))
        return anext((elem async for elem in iterable if pred(elem) == v), None)

    converted = [(attrgetter(attr.replace("__", ".")), value) for attr, value in attrs.items()]

    return anext((elem async for elem in iterable if all(pred(elem) == value for pred, value in converted)), None)


@overload
def get(iterable: AsyncIterable[_T], /, **attrs: Any) -> Coro[_T | None]: ...


@overload
def get(iterable: Iterable[_T], /, **attrs: Any) -> _T | None: ...


def get(iterable: _Iter[_T], /, **attrs: Any) -> _T | None | Coro[_T | None]:
    """A helper that returns the first element in the iterable that meets all the traits passed in ``attrs``. This
    is an alternative for :func:`find`.

    Examples
    --------
    .. code:: python

        bff = steam.utils.get(client.users, name="Gobot1234")

        trade = steam.utils.get(client.trades, state=TradeOfferState.Active, partner=message.author)
        # multiple attributes are also accepted

    Parameters
    -----------
    iterable
        An iterable to search through.
    attrs
        Keyword arguments that denote attributes to match.

    Returns
    -------
    The first element from the ``iterable`` which matches all the traits passed in ``attrs`` or ``None`` if no matching
    element was found.
    """
    return (
        _aget(iterable, **attrs)  # type: ignore
        if hasattr(iterable, "__aiter__")  # isinstance(iterable, collections.abc.AsyncIterable) is too slow
        else _get(iterable, **attrs)  # type: ignore
    )


async def maybe_coroutine(
    func: Callable[_P, _T | Awaitable[_T]],
    /,
    *args: _P.args,
    **kwargs: _P.kwargs,
) -> _T:
    value = func(*args, **kwargs)
    return await value if isawaitable(value) else value



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

import string
from dataclasses import dataclass
from typing import TYPE_CHECKING

from yarl import URL as URL_

from ._const import URL
from .app import PartialApp
from .bundle import PartialBundle
from .clan import PartialClan
from .enums import AppType, ContentDescriptor, Currency, Language, PaymentMethod, ReviewType
from .models import CDNAsset
from .package import PartialPackage
from .protobufs import store
from .tag import Category, Tag
from .utils import DateTime

if TYPE_CHECKING:
    from datetime import datetime

    from .state import ConnectionState

__all__ = (
    "StoreItem",
    "AppStoreItem",
    "PackageStoreItem",
    "BundleStoreItem",
    "TransactionReceipt",
)


class StoreItemTag(Tag[None]):
    __slots__ = ("weight",)

    def __init__(self, state: ConnectionState, id: int, weight: int):
        super().__init__(state, id)
        self.weight = weight


@dataclass(slots=True)
class StoreItemCreatorHomeLink:
    name: str
    clan: PartialClan


@dataclass(slots=True)
class StoreItemPurchaseOptionDiscount:
    amount: int
    description: str
    ends_at: datetime


class StoreItemPurchaseOption:
    __slots__ = (
        "package",
        "bundle",
        "final_price_in_cents",
        "original_price_in_cents",
        "user_final_price_in_cents",
        "formatted_final_price",
        "formatted_original_price",
        "discount_percentage",
        "user_discount_percentage",
        "bundle_discount_percentage",
        "active_discounts",
        "user_active_discounts",
        "inactive_discounts",
        "user_can_purchase",
        "user_can_purchase_as_gift",
        "is_commercial_license",
        "should_suppress_discount_percentage",
    )

    def __init__(self, state: ConnectionState, proto: store.StoreItemPurchaseOption) -> None:
        self.package = PartialPackage(state, id=proto.packageid) if proto.packageid else None
        self.bundle = PartialBundle(state, id=proto.bundleid) if proto.bundleid else None
        self.final_price_in_cents: int = proto.final_price_in_cents
        self.original_price_in_cents: int = proto.original_price_in_cents
        self.user_final_price_in_cents: int = proto.user_final_price_in_cents
        self.formatted_final_price: str = proto.formatted_final_price
        self.formatted_original_price: str = proto.formatted_original_price
        self.discount_percentage: int = proto.discount_pct
        self.user_discount_percentage: int = proto.user_discount_pct
        self.bundle_discount_percentage: int = proto.bundle_discount_pct
        self.active_discounts: list[StoreItemPurchaseOptionDiscount] = [
            StoreItemPurchaseOptionDiscount(
                discount.discount_amount,
                discount.discount_description,
                DateTime.from_timestamp(discount.discount_end_date),
            )
            for discount in proto.active_discounts
        ]
        self.user_active_discounts: list[StoreItemPurchaseOptionDiscount] = [
            StoreItemPurchaseOptionDiscount(
                discount.discount_amount,
                discount.discount_description,
                DateTime.from_timestamp(discount.discount_end_date),
            )
            for discount in proto.user_active_discounts
        ]
        self.inactive_discounts: list[StoreItemPurchaseOptionDiscount] = [
            StoreItemPurchaseOptionDiscount(
                discount.discount_amount,
                discount.discount_description,
                DateTime.from_timestamp(discount.discount_end_date),
            )
            for discount in proto.inactive_discounts
        ]
        self.user_can_purchase: bool = proto.user_can_purchase
        self.user_can_purchase_as_gift: bool = proto.user_can_purchase_as_gift
        self.is_commercial_license: bool = proto.is_commercial_license
        self.should_suppress_discount_percentage: bool = proto.should_suppress_discount_pct


@dataclass(slots=True)
class LanguageSupport:
    language: Language
    supported: bool
    audio: bool
    subtitles: bool


class StoreItem:
    __slots__ = SLOTS = (  # done to avoid "multiple bases have instance lay-out conflict"
        "_language",
        "hidden",
        "unavailable_for_country_restriction",
        "_apps",
        "created_at",
        "type",
        "included_types",
        "related_parent_app",
        "tags",
        "categories",
        "review_score",
        "review_count",
        "review_percent_positive",
        "filtered_review_score",
        "filtered_review_count",
        "filtered_review_percent_positive",
        "summary",
        "capsule_headline",
        "publishers",
        "developers",
        "franchises",
        "small_capsule",
        "header",
        "package_header",
        "page_background",
        "hero_capsule",
        "library_capsule",
        "library_hero",
        "community_icon",
        "vr_support",
        "game_rating",
        "best_purchase_option",
        "purchase_options",
        "accessories",
        "all_ages_screenshots",
        "mature_screenshots",
        "trailers",
        "supported_languages",
        "free_weekend",
        "content_descriptors",
        "_free",
        "_early_access",
        "_on_windows",
        "_on_mac_os",
        "_on_linux",
    )
    if not TYPE_CHECKING:
        __slots__ = ()

    def __init__(self, state: ConnectionState, proto: store.StoreItem, language: Language) -> None:
        super().__init__(state, id=proto.id, name=proto.name)  # type: ignore  # this is a mixin
        self._language = language
        self.hidden = not proto.visible
        self.unavailable_for_country_restriction = proto.unvailable_for_country_restriction
        self._apps = [PartialApp(state, id=id) for id in proto.included_appids]
        self.created_at = DateTime.from_timestamp(proto.release.steam_release_date)
        proto.type = store.EStoreAppType(proto.type)  # TODO remove when enum redux pr is merged
        try:
            self.type = AppType[proto.type.name]
        except KeyError:
            self.type = AppType.try_value(proto.type)  # should be good enough
        self.included_types: list[AppType] = []
        for included_type in proto.included_types:
            included_type = store.EStoreAppType(included_type)  # TODO remove when enum redux pr is merged
            try:
                self.included_types.append(AppType[included_type.name])
            except KeyError:
                self.included_types.append(AppType.try_value(included_type))

        self.related_parent_app = PartialApp(state, id=proto.related_items.parent_appid)
        self.tags = [StoreItemTag(state, tag.tagid, tag.weight) for tag in proto.tags]
        self.categories = [Category(state, id) for id in proto.categories.feature_categoryids]
        self.review_score = ReviewType.try_value(proto.reviews.summary_unfiltered.review_score)
        self.review_count = proto.reviews.summary_unfiltered.review_count
        self.review_percent_positive = proto.reviews.summary_unfiltered.percent_positive
        self.filtered_review_score = ReviewType.try_value(proto.reviews.summary_filtered.review_score)
        self.filtered_review_count = proto.reviews.summary_filtered.review_count
        self.filtered_review_percent_positive = proto.reviews.summary_filtered.percent_positive

        self.summary = proto.basic_info.short_description
        self.capsule_headline = proto.basic_info.capsule_headline
        self.publishers = [
            StoreItemCreatorHomeLink(name=link.name, clan=PartialClan(state, id=link.creator_clan_account_id))
            for link in proto.basic_info.publishers
        ]
        self.developers = [
            StoreItemCreatorHomeLink(name=link.name, clan=PartialClan(state, id=link.creator_clan_account_id))
            for link in proto.basic_info.developers
        ]
        self.franchises = [
            StoreItemCreatorHomeLink(name=link.name, clan=PartialClan(state, id=link.creator_clan_account_id))
            for link in proto.basic_info.franchises
        ]

        base_url = string.Template(f"{URL.CDN}/{proto.assets.asset_url_format}")
        capsule = proto.assets.main_capsule or proto.assets.small_capsule or None
        self.small_capsule = CDNAsset(state, base_url.substitute(FILENAME=capsule)) if capsule else None
        self.header = (
            CDNAsset(state, base_url.substitute(FILENAME=proto.assets.header)) if proto.assets.header else None
        )
        self.package_header = (
            CDNAsset(state, base_url.substitute(FILENAME=proto.assets.package_header))
            if proto.assets.package_header
            else None
        )
        self.page_background = (
            CDNAsset(state, base_url.substitute(FILENAME=proto.assets.page_background))
            if proto.assets.page_background
            else None
        )
        hero_capsule = proto.assets.hero_capsule_2_x or proto.assets.hero_capsule or None
        self.hero_capsule = CDNAsset(state, base_url.substitute(FILENAME=hero_capsule)) if hero_capsule else None
        library_capsule = proto.assets.library_capsule_2_x or proto.assets.library_capsule or None
        self.library_capsule = (
            CDNAsset(state, base_url.substitute(FILENAME=library_capsule)) if library_capsule else None
        )
        library_hero = proto.assets.library_hero_2_x or proto.assets.library_hero or None
        self.library_hero = CDNAsset(state, base_url.substitute(FILENAME=library_hero)) if library_hero else None
        community_icon = URL_(base_url.substitute(FILENAME=proto.assets.community_icon))
        self.community_icon = (
            CDNAsset(state, str(community_icon) if community_icon.suffix else f"{community_icon}.jpg")
            if proto.assets.community_icon
            else None
        )  # this doesn't seem to have an extension in the normal case

        self.vr_support = proto.platforms.vr_support
        self.game_rating = proto.game_rating

        self.best_purchase_option = StoreItemPurchaseOption(state, proto.best_purchase_option)
        self.purchase_options = [
            StoreItemPurchaseOption(state, purchase_option) for purchase_option in proto.purchase_options
        ]
        self.accessories = [StoreItemPurchaseOption(state, purchase_option) for purchase_option in proto.accessories]
        self.all_ages_screenshots = [
            CDNAsset(state, str(URL.CDN / proto.filename)) for proto in proto.screenshots.all_ages_screenshots
        ]
        self.mature_screenshots = [
            CDNAsset(state, str(URL.CDN / proto.filename)) for proto in proto.screenshots.mature_content_screenshots
        ]
        self.trailers = proto.trailers
        self.supported_languages = [
            LanguageSupport(Language.try_value(proto.elanguage), proto.supported, proto.full_audio, proto.subtitles)
            for proto in proto.supported_languages
        ]
        self.free_weekend = proto.free_weekend or None

        self.content_descriptors = [ContentDescriptor.try_value(d) for d in proto.content_descriptorids]

        self._free = proto.is_free
        self._early_access = proto.is_early_access
        self._on_windows = proto.platforms.windows
        self._on_mac_os = proto.platforms.mac
        self._on_linux = proto.platforms.steamos_linux

    def is_free(self) -> bool:
        """Whether the store item is free to play."""
        return self._free

    def is_early_access(self) -> bool:
        """Whether the store item is in early access."""
        return self._early_access

    def is_on_windows(self) -> bool:
        """Whether the store item is playable on Windows."""
        return self._on_windows

    def is_on_mac_os(self) -> bool:
        """Whether the store item is playable on macOS."""
        return self._on_mac_os

    def is_on_linux(self) -> bool:
        """Whether the store item is playable on Linux."""
        return self._on_linux


class AppStoreItem(StoreItem, PartialApp[str]):
    """Represents an app fetched from the store."""

    __slots__ = StoreItem.SLOTS


class PackageStoreItem(StoreItem, PartialPackage[str]):
    """Represents a package fetched from the store."""

    __slots__ = StoreItem.SLOTS


class BundleStoreItem(StoreItem, PartialBundle[str]):
    """Represents a bundle fetched from the store."""

    __slots__ = StoreItem.SLOTS


@dataclass(slots=True)
class TransactionReceiptItem:
    package: PartialPackage[str] | None
    """The package that was purchased."""
    app: PartialApp[str] | None
    """The app that was purchased."""


class TransactionReceipt:
    """Represents a transaction receipt."""

    def __init__(self, state: ConnectionState, proto: store.PurchaseReceiptInfo):
        self.id = proto.transactionid
        """The ID of the transaction."""
        self.items = [
            TransactionReceiptItem(
                PartialPackage(state, id=item.packageid, name=item.line_item_description) if item.packageid else None,
                PartialApp(state, id=item.appid, name=item.line_item_description) if item.appid else None,
            )
            for item in proto.line_items
        ]  # TODO research if this is a union?
        """The items in the transaction."""

        self.package = PartialPackage(state, id=proto.packageid)  # TODO can this be upgraded to a License?
        """The package in the transaction."""
        self.purchase_status = proto.purchase_status
        """The status of the transaction."""
        self.created_at = DateTime.from_timestamp(proto.transaction_time)
        """The time the transaction was created."""
        self.payment_method = PaymentMethod.try_value(proto.payment_method)
        """The payment method used for the transaction."""
        self.base_price = proto.base_price
        """The base price of the transaction."""
        self.total_discount = proto.total_discount
        """The total discount of the transaction."""
        self.tax = proto.tax
        """The tax of the transaction."""
        self.shipping = proto.shipping
        """The shipping cost of the transaction."""
        self.currency = Currency.try_value(proto.currency_code)
        """The currency code of the transaction."""
        self.language = Language.from_web_api_str(proto.country_code)
        """The country code of the transaction."""



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: engine_gcmessages.proto
# plugin: python-betterproto

from dataclasses import dataclass

import betterproto


@dataclass(eq=False, repr=False)
class GotvSyncPacket(betterproto.Message):
    match_id: int = betterproto.uint64_field(1)
    instance_id: int = betterproto.uint32_field(2)
    signupfragment: int = betterproto.uint32_field(3)
    currentfragment: int = betterproto.uint32_field(4)
    tickrate: float = betterproto.float_field(5)
    tick: int = betterproto.uint32_field(6)
    rtdelay: float = betterproto.float_field(8)
    rcvage: float = betterproto.float_field(9)
    keyframe_interval: float = betterproto.float_field(10)
    cdndelay: int = betterproto.uint32_field(11)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: gcsystemmsgs.proto
# plugin: python-betterproto

from dataclasses import dataclass

import betterproto


@dataclass(eq=False, repr=False)
class CMsgGchVacVerificationChange(betterproto.Message):
    steamid: int = betterproto.fixed64_field(1)
    appid: int = betterproto.uint32_field(2)
    is_verified: bool = betterproto.bool_field(3)


@dataclass(eq=False, repr=False)
class CMsgGchAccountPhoneNumberChange(betterproto.Message):
    steamid: int = betterproto.fixed64_field(1)
    appid: int = betterproto.uint32_field(2)
    phone_id: int = betterproto.uint64_field(3)
    is_verified: bool = betterproto.bool_field(4)
    is_identifying: bool = betterproto.bool_field(5)


@dataclass(eq=False, repr=False)
class CMsgGchInviteUserToLobby(betterproto.Message):
    steamid: int = betterproto.fixed64_field(1)
    appid: int = betterproto.uint32_field(2)
    steamid_invited: int = betterproto.fixed64_field(3)
    steamid_lobby: int = betterproto.fixed64_field(4)


@dataclass(eq=False, repr=False)
class CQuestPublisherAddCommunityItemsToPlayerRequest(betterproto.Message):
    steamid: int = betterproto.uint64_field(1)
    appid: int = betterproto.uint32_field(2)
    match_item_type: int = betterproto.uint32_field(3)
    match_item_class: int = betterproto.uint32_field(4)
    prefix_item_name: str = betterproto.string_field(5)
    attributes: "list[CQuestPublisherAddCommunityItemsToPlayerRequestAttribute]" = betterproto.message_field(6)
    note: str = betterproto.string_field(7)


@dataclass(eq=False, repr=False)
class CQuestPublisherAddCommunityItemsToPlayerRequestAttribute(betterproto.Message):
    attribute: int = betterproto.uint32_field(1)
    value: int = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class CQuestPublisherAddCommunityItemsToPlayerResponse(betterproto.Message):
    items_matched: int = betterproto.uint32_field(1)
    items_granted: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class CCommunityGamePersonalDataCategoryInfo(betterproto.Message):
    type: str = betterproto.string_field(1)
    localization_token: str = betterproto.string_field(2)
    template_file: str = betterproto.string_field(3)


@dataclass(eq=False, repr=False)
class CCommunityGetGamePersonalDataCategoriesRequest(betterproto.Message):
    appid: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class CCommunityGetGamePersonalDataCategoriesResponse(betterproto.Message):
    categories: "list[CCommunityGamePersonalDataCategoryInfo]" = betterproto.message_field(1)
    app_assets_basename: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class CCommunityGetGamePersonalDataEntriesRequest(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    steamid: int = betterproto.uint64_field(2)
    type: str = betterproto.string_field(3)
    continue_token: str = betterproto.string_field(4)


@dataclass(eq=False, repr=False)
class CCommunityGetGamePersonalDataEntriesResponse(betterproto.Message):
    gceresult: int = betterproto.uint32_field(1)
    entries: list[str] = betterproto.string_field(2)
    continue_token: str = betterproto.string_field(3)


@dataclass(eq=False, repr=False)
class CCommunityTerminateGamePersonalDataEntriesRequest(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    steamid: int = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class CCommunityTerminateGamePersonalDataEntriesResponse(betterproto.Message):
    gceresult: int = betterproto.uint32_field(1)



from __future__ import annotations

from typing import TYPE_CHECKING

from ...enums import Enum, Flags, IntEnum, classproperty

if TYPE_CHECKING:
    from typing_extensions import Self

__all__ = (
    "GCGoodbyeReason",
    "TradeResponse",
    "Mercenary",
    "ItemSlot",
    "WearLevel",
    "BackpackSortType",
    "ItemFlags",
    "ItemOrigin",
    "ItemQuality",
    "Part",
    "Spell",
    "Sheen",
    "Killstreak",
    "Attribute",
    "EMsg",
)


# fmt: off
class GCGoodbyeReason(IntEnum):
    GCGoingDown = 1
    NoSession   = 2


class TradeResponse(IntEnum):
    Accepted                        = 0
    Declined                        = 1
    TradeBannedInitiator            = 2
    TradeBannedTarget               = 3
    TargetAlreadyTrading            = 4
    Disabled                        = 5
    NotLoggedIn                     = 6
    Cancel                          = 7
    TooSoon                         = 8
    TooSoonPenalty                  = 9
    ConnectionFailed                = 10
    AlreadyTrading                  = 11
    AlreadyHasTradeRequest          = 12
    NoResponse                      = 13
    CyberCafeInitiator              = 14
    CyberCafeTarget                 = 15
    SchoolLabInitiator              = 16
    SchoolLabTarget                 = 16
    InitiatorBlockedTarget          = 18
    InitiatorNeedsVerifiedEmail     = 20
    InitiatorNeedsSteamGuard        = 21
    TargetAccountCannotTrade        = 22
    InitiatorSteamGuardDuration     = 23
    InitiatorPasswordResetProbation = 24
    InitiatorNewDeviceCooldown      = 25
    OKToDeliver                     = 50


class Mercenary(IntEnum):
    Scout    = 1
    Sniper   = 2
    Soldier  = 3
    Demoman  = 4
    Medic    = 5
    Heavy    = 6
    Pyro     = 7
    Spy      = 8
    Engineer = 9


class ItemSlot(IntEnum):
    Primary   = 0
    Secondary = 1
    Melee     = 2
    Sapper    = 4
    PDA       = 5
    PDA2      = 6
    Cosmetic1 = 7
    Cosmetic2 = 8
    Action    = 9
    Cosmetic3 = 10
    Taunt1    = 11
    Taunt2    = 12
    Taunt3    = 13
    Taunt4    = 14
    Taunt5    = 15
    Taunt6    = 16
    Taunt7    = 17
    Taunt8    = 18
    Misc      = 100  # these are not real but are used for BackPackItem.slot
    Gift      = 101
    CraftItem = 102


class WearLevel(Enum):
    FactoryNew    = "(Factory New)"
    MinimalWear   = "(Minimal Wear)"
    FieldTested   = "(Field-Tested)"
    WellWorn      = "(Well-Worn)"
    BattleScarred = "(Battle Scarred)"


class BackpackSortType(IntEnum):  # N.B only in game ones will actually work
    Name     = 1
    Defindex = 2
    Rarity   = 3
    Type     = 4
    Date     = 5
    Class    = 101
    Slot     = 102


# lifted from https://github.com/OthmanAba/TeamFortress2/blob/master/tf2_src/game/shared/econ/econ_item_constants.h
class ItemFlags(Flags):
    CannotTrade                                = 1 << 0
    CannotCraft                                = 1 << 1
    CanBeTradedByFreeAccounts                  = 1 << 2
    NotEcon                                    = 1 << 3
    """Items that cannot interact in the economy (can't be traded, gift-wrapped, crafted, etc.)"""
    PurchasedAfterStoreCraftabilityChanges2012 = 1 << 4
    """Cosmetic items coming from the store are now usable in crafting"""
    ForceBlueTeam                              = 1 << 5
    StoreItem                                  = 1 << 6
    Preview                                    = 1 << 7


class ItemOrigin(IntEnum):
    Invalid                        = -1
    Drop                           = 0
    Achievement                    = 1
    Purchased                      = 2
    Traded                         = 3
    Crafted                        = 4
    StorePromotion                 = 5
    Gifted                         = 6
    SupportGranted                 = 7
    FoundInCrate                   = 8
    Earned                         = 9
    ThirdPartyPromotion            = 10
    GiftWrapped                    = 11
    HalloweenDrop                  = 12
    PackageItem                    = 13
    Foreign                        = 14
    CDKey                          = 15
    CollectionReward               = 16
    PreviewItem                    = 17
    SteamWorkshopContribution      = 18
    PeriodicScoreReward            = 19
    MvMMissionCompletionReward     = 20
    MvMSquadSurplusReward          = 21
    RecipeOutput                   = 22
    QuestDrop                      = 23
    QuestLoanerItem                = 24
    TradeUp                        = 25
    ViralCompetitiveBetaPassSpread = 26
    Max                            = 27


class ItemQuality(IntEnum):
    Normal          = 0
    Genuine         = 1
    Vintage         = 3
    Rarity3         = 4
    Unusual         = 5
    Unique          = 6
    Community       = 7
    Valve           = 8
    SelfMade        = 9
    Customized      = 10
    Strange         = 11
    Completed       = 12
    Haunted         = 13
    Collectors      = 14
    DecoratedWeapon = 15


# class Paint(IntEnum):
#     An Air of Debonair = 0x654740
#     An Air of Debonair = 0x28394D
#     Balaclavas Are Forever = 0x3B1F23
#     Balaclavas Are Forever = 0x18233D
#     Cream Spirit = 0xC36C2D
#     Cream Spirit = 0xB88035
#     Operator's Overalls = 0x483838
#     Operator's Overalls = 0x384248
#     Team Spirit = 0xB8383B
#     Team Spirit = 0x5885A2
#     The Value of Teamwork = 0x803020
#     The Value of Teamwork = 0x256D8D
#     Waterlogged Lab Coat = 0xA89A8C
#     Waterlogged Lab Coat = 0x839FA3
#     A Color Similar to Slate = 0x2F4F4F
#     A Deep Commitment to Purple = 0x7D4071
#     A Distinctive Lack of Hue = 0x141414
#     A Mann's Mint = 0xBCDDB3
#     After Eight = 0x2D2D24
#     Aged Moustache Grey = 0x7E7E7E
#     An Extraordinary Abundance of Tinge = 0xE6E6E6
#     Australium Gold = 0xE7B53B
#     Color No. 216-190-216 = 0xD8BED8
#     Dark Salmon Injustice = 0xE9967A
#     Drably Olive = 0x808000
#     Indubitably Green = 0x729E42
#     Mann Co. Orange = 0xCF7336
#     Muskelmannbraun = 0xA57545
#     Noble Hatter's Violet = 0x51384A
#     Peculiarly Drab Tincture = 0xC5AF91
#     Pink as Hell = 0xFF69B4
#     Radigan Conagher Brown = 0x694D3A
#     The Bitter Taste of Defeat and Lime = 0x32CD32
#     The Color of a Gentlemann's Business Pants = 0xF0E68C
#     Ye Olde Rustic Colour = 0x7C6C57
#     Zepheniah's Gree = 0x424F3B


class Part(IntEnum):
    # Kills                              = 0
    # Ubers                              = 1
    KillAssists                        = 2
    # SentryKills                        = 3
    # SoddenVictims                      = 4
    # SpiesShocked                       = 5
    # HeadsTaken                         = 6
    # Humiliations                       = 7
    # GiftsGiven                         = 8
    # DeathsFeigned                      = 9
    ScoutsKilled                       = 10
    SnipersKilled                      = 11
    SoldiersKilled                     = 12
    DemomenKilled                      = 13
    HeaviesKilled                      = 14
    PyrosKilled                        = 15
    SpiesKilled                        = 16
    EngineersKilled                    = 17
    MedicsKilled                       = 18
    BuildingsDestroyed                 = 19
    ProjectilesReflected               = 20
    HeadshotKills                      = 21
    AirborneEnemyKills                 = 22
    GibKills                           = 23
    # BuildingsSapped                    = 24
    # TickleFightsWon                    = 25
    # OpponentsFlattened                 = 26
    KillsUnderAFullMoon                = 27
    Dominations                        = 28
    Revenges                           = 30
    PosthumousKills                    = 31
    TeammatesExtinguished              = 32
    CriticalKills                      = 33
    KillsWhileExplosiveJumping         = 34
    SappersDestroyed                   = 36
    CloakedSpiesKilled                 = 37
    MedicsKilledThatHaveFullÜberCharge = 38
    RobotsDestroyed                    = 39
    GiantRobotsDestroyed               = 40
    KillsWhileLowHealth                = 44
    KillsDuringHalloween               = 45
    RobotsKilledDuringHalloween        = 46
    DefenderKills                      = 47
    SubmergedEnemyKills                = 48
    # KillsWhileInvulnÜberCharged        = 49
    # FoodItemsEaten                     = 50
    # BannersDeployed                    = 51
    # SecondsCloaked                     = 58
    # HealthDispensedtoTeammates         = 59
    # TeammatesTeleported                = 60
    TanksDestroyed                     = 61
    LongDistanceKills                  = 62
    # KillEaterEvent_UniquePlayerKills   = 63
    # PointsScored                       = 64
    # DoubleDonks                        = 65
    # TeammatesWhipped                   = 66
    KillsduringVictoryTime             = 67
    RobotScoutsDestroyed               = 68
    RobotSpiesDestroyed                = 74
    TauntKills                         = 77
    UnusualWearingPlayerKills          = 78
    BurningPlayerKills                 = 79
    KillstreaksEnded                   = 80
    FreezecamTauntAppearances          = 81
    DamageDealt                        = 82
    FiresSurvived                      = 83
    AlliedHealingDone                  = 84
    PointBlankKills                    = 85
    # WrangledSentryKills                = 86
    Kills                              = 87
    FullHealthKills                    = 88
    TauntingPlayerKills                = 89
    # CarnivalKills                      = 90
    # CarnivalUnderworldKills            = 91
    # CarnivalGamesWon                   = 92
    NotCritnorMiniCritKills            = 93
    PlayerHits                         = 94
    Assists                            = 95
    # ContractsCompleted                 = 96
    Kills_                             = 97
    # ContractPoints                     = 98
    # ContractBonusPoints                = 99
    # TimesPerformed                     = 100
    # KillsAndAssistsduringInvasionEvent = 101
    # KillsAndAssistson2FortInvasion     = 102
    # KillsAndAssistsonProbed            = 103
    # KillsAndAssistsonByre              = 104
    # KillsAndAssistsonWatergate         = 105
    # SoulsCollected                     = 106
    # MerasmissionsCompleted             = 107
    # HalloweenTransmutesPerformed       = 108
    # PowerUpCanteensUsed                = 109
    # ContractPointsEarned               = 110
    # ContractPointsContributedToFriends = 111

class Spell(IntEnum):
    VoicesFromBelow         = 1006
    PumpkinBombs            = 1007
    HalloweenFire           = 1008
    Exorcism                = 1009
    PutrescentPigmentation  = 8900
    DieJob                  = 8901
    ChromaticCorruption     = 8902
    SpectralSpectrum        = 8903
    SinisterStaining        = 8904
    TeamSpiritFootprints    = 8914
    GangreenFootprints      = 8915
    CorpseGrayFootprints    = 8916
    ViolentVioletFootprints = 8917
    RottenOrangeFootprints  = 8918
    BruisedPurpleFootprints = 8919
    HeadlessHorseshoe       = 8920

    @classproperty
    def indices(cls: type[Self]) -> dict[int, dict[int, Spell]]:  # type: ignore
        return {
            1004: {
                0: cls.DieJob,
                1: cls.ChromaticCorruption,
                2: cls.PutrescentPigmentation,
                3: cls.SpectralSpectrum,
                4: cls.SinisterStaining,
            },
            1005: {
                1: cls.TeamSpiritFootprints,
                2: cls.HeadlessHorseshoe,
                3100495: cls.CorpseGrayFootprints,
                5322826: cls.ViolentVioletFootprints,
                8208497: cls.BruisedPurpleFootprints,
                8421376: cls.GangreenFootprints,
                13595446: cls.RottenOrangeFootprints,
            }
        }


class Sheen(IntEnum):
    TeamShine        = 1
    DeadlyDaffodil   = 2
    Manndarin        = 3
    MeanGreen        = 4
    AgonizingEmerald = 5
    VillainousViolet = 6
    HotRod           = 7


class Killstreak(IntEnum):
    FireHorns         = 2002
    CerebralDischarge = 2003
    Tornado           = 2004
    Flames            = 2005
    Singularity       = 2006
    Incinerator       = 2007
    HypnoBeam         = 2008


class Attribute(IntEnum):
    Paint         = 1031
    CustomTexture = 1051
    MakersMark    = 1053
    Killstreak    = 1094
    GiftedBy      = 2570
    Festivizer    = 2572


class EMsg(IntEnum):
    SOCreate                                    = 21
    SOUpdate                                    = 22
    SODestroy                                   = 23
    SOCacheSubscribed                           = 24
    SOCacheUnsubscribed                         = 25
    SOUpdateMultiple                            = 26
    SOCacheSubscriptionCheck                    = 27
    SOCacheSubscriptionRefresh                  = 28
    SOCacheSubscribedUpToDate                   = 29

    Base                                        = 1000
    SetSingleItemPosition                       = 1001
    Craft                                       = 1002
    CraftResponse                               = 1003
    Delete                                      = 1004
    VerifyCacheSubscription                     = 1005
    NameItem                                    = 1006
    UnlockCrate                                 = 1007
    UnlockCrateResponse                         = 1008
    PaintItem                                   = 1009
    PaintItemResponse                           = 1010
    GoldenWrenchBroadcast                       = 1011
    MOTDRequest                                 = 1012
    MOTDRequestResponse                         = 1013
    NameBaseItem                                = 1019
    NameBaseItemResponse                        = 1020
    CustomizeItemTexture                        = 1023
    CustomizeItemTextureResponse                = 1024
    UseItemRequest                              = 1025
    UseItemResponse                             = 1026
    RespawnPostLoadoutChange                    = 1029
    RemoveItemName                              = 1030
    RemoveItemPaint                             = 1031
    GiftWrapItem                                = 1032
    GiftWrapItemResponse                        = 1033
    DeliverGift                                 = 1034
    DeliverGiftResponseReceiver                 = 1036
    UnwrapGiftRequest                           = 1037
    UnwrapGiftResponse                          = 1038
    SetItemStyle                                = 1039
    UsedClaimCodeItem                           = 1040
    SortItems                                   = 1041
    LookupAccount                               = 1043
    LookupAccountResponse                       = 1044
    LookupAccountName                           = 1045
    LookupAccountNameResponse                   = 1046
    UpdateItemSchema                            = 1049
    RequestInventoryRefresh                     = 1050
    RemoveCustomTexture                         = 1051
    RemoveCustomTextureResponse                 = 1052
    RemoveMakersMark                            = 1053
    RemoveMakersMarkResponse                    = 1054
    RemoveUniqueCraftIndex                      = 1055
    RemoveUniqueCraftIndexResponse              = 1056
    SaxxyBroadcast                              = 1057
    BackpackSortFinished                        = 1058
    AdjustItemEquippedState                     = 1059
    CollectItem                                 = 1061
    ItemAcknowledged                            = 1062
    PresetsSelectPresetForClass                 = 1063
    PresetsSetItemPosition                      = 1064
    ReportAbuse                                 = 1065
    ReportAbuseResponse                         = 1066
    PresetsSelectPresetForClassReply            = 1067
    NameItemNotification                        = 1068
    ClientDisplayNotification                   = 1069
    ApplyStrangePart                            = 1070
    IncrementKillCountAttribute                 = 1071
    IncrementKillCountResponse                  = 1072
    RemoveStrangePart                           = 1073
    ResetStrangeScores                          = 1074
    GiftedItems                                 = 1075
    ApplyUpgradeCard                            = 1077
    RemoveUpgradeCard                           = 1078
    ApplyStrangeRestriction                     = 1079
    ClientRequestMarketData                     = 1080
    ClientRequestMarketDataResponse             = 1081
    ApplyXifier                                 = 1082
    ApplyXifierResponse                         = 1083
    TrackUniquePlayerPairEvent                  = 1084
    FulfillDynamicRecipeComponent               = 1085
    FulfillDynamicRecipeComponentResponse       = 1086
    SetItemEffectVerticalOffset                 = 1087
    SetHatEffectUseHeadOrigin                   = 1088
    ItemEaterRecharger                          = 1089
    ItemEaterRechargerResponse                  = 1090
    ApplyBaseItemXifier                         = 1091
    ApplyClassTransmogrifier                    = 1092
    ApplyHalloweenSpellbookPage                 = 1093
    RemoveKillStreak                            = 1094
    RemoveKillStreakResponse                    = 1095
    TFSpecificItemBroadcast                     = 1096
    IncrementKillCountAttributeMultiple         = 1097
    DeliverGiftResponseGiver                    = 1098
    SetItemPositions                            = 1100
    LookupMultipleAccountNames                  = 1101
    LookupMultipleAccountNamesResponse          = 1102
    TradingBase                                 = 1500
    TradingInitiateTradeRequest                 = 1501
    TradingInitiateTradeResponse                = 1502
    TradingStartSession                         = 1503
    TradingSessionClosed                        = 1509
    TradingCancelSession                        = 1510
    TradingInitiateTradeRequestResponse         = 1514
    ServerBrowserFavoriteServer                 = 1601
    ServerBrowserBlacklistServer                = 1602
    ServerRentalsBase                           = 1700
    ItemPreviewCheckStatus                      = 1701
    ItemPreviewStatusResponse                   = 1702
    ItemPreviewRequest                          = 1703
    ItemPreviewRequestResponse                  = 1704
    ItemPreviewExpire                           = 1705
    ItemPreviewExpireNotification               = 1706
    ItemPreviewItemBoughtNotification           = 1708
    DevNewItemRequest                           = 2001
    DevNewItemRequestResponse                   = 2002
    DevDebugRollLootRequest                     = 2003

    StoreGetUserData                            = 2500
    StoreGetUserDataResponse                    = 2501
    StorePurchaseFinalize                       = 2512
    StorePurchaseFinalizeResponse               = 2513
    StorePurchaseCancel                         = 2514
    StorePurchaseCancelResponse                 = 2515
    StorePurchaseQueryTxn                       = 2508
    StorePurchaseQueryTxnResponse               = 2509
    StorePurchaseInit                           = 2510
    StorePurchaseInitResponse                   = 2511
    GCToGCDirtySDOCache                         = 2516
    GCToGCDirtyMultipleSDOCache                 = 2517
    GCToGCUpdateSQLKeyValue                     = 2518
    GCToGCBroadcastConsoleCommand               = 2521
    ServerVersionUpdated                        = 2522
    ApplyAutograph                              = 2523
    GCToGCWebAPIAccountChanged                  = 2524
    RequestAnnouncements                        = 2525
    RequestAnnouncementsResponse                = 2526
    RequestPassportItemGrant                    = 2527
    ClientVersionUpdated                        = 2528
    ItemPurgatoryFinalizePurchase               = 2531
    ItemPurgatoryFinalizePurchaseResponse       = 2532
    ItemPurgatoryRefundPurchase                 = 2533
    ItemPurgatoryRefundPurchaseResponse         = 2534
    GCToGCPlayerStrangeCountAdjustments         = 2535
    RequestStoreSalesData                       = 2536
    RequestStoreSalesDataResponse               = 2537
    RequestStoreSalesDataUpToDateResponse       = 2538
    GCToGCPingRequest                           = 2539
    GCToGCPingResponse                          = 2540
    GCToGCGetUserSessionServer                  = 2541
    GCToGCGetUserSessionServerResponse          = 2542
    GCToGCGetUserServerMembers                  = 2543
    GCToGCGetUserServerMembersResponse          = 2544
    GCToGCGrantSelfMadeItemToAccount            = 2555
    GCToGCThankedByNewUser                      = 2556
    ShuffleCrateContents                        = 2557
    QuestObjectiveProgress                      = 2558
    QuestCompleted                              = 2559
    ApplyDuckToken                              = 2560
    QuestCompleteRequest                        = 2561
    QuestObjectivePointsChange                  = 2562
    QuestObjectiveRequestLoanerItems            = 2564
    QuestObjectiveRequestLoanerResponse         = 2565
    ApplyStrangeCountTransfer                   = 2566
    CraftCollectionUpgrade                      = 2567
    CraftHalloweenOffering                      = 2568
    QuestDiscardRequest                         = 2569
    RemoveGiftedBy                              = 2570
    RemoveGiftedByResponse                      = 2571
    RemoveFestivizer                            = 2572
    RemoveFestivizerResponse                    = 2573
    CraftCommonStatClock                        = 2574


    PingRequest                                 = 3001
    PingResponse                                = 3002

    ClientWelcome                               = 4004
    ServerWelcome                               = 4005
    ClientHello                                 = 4006
    ServerHello                                 = 4007
    ClientGoodbye                               = 4008
    ServerGoodbye                               = 4009

    SystemMessage                               = 4001
    ReplicateConVars                            = 4002
    ConVarUpdated                               = 4003
    InviteToParty                               = 4501
    InvitationCreated                           = 4502
    PartyInviteResponse                         = 4503
    KickFromParty                               = 4504
    LeaveParty                                  = 4505
    ServerAvailable                             = 4506
    ClientConnectToServer                       = 4507
    GameServerInfo                              = 4508
    Error                                       = 4509
    ReplayUploadedToYouTube                     = 4510
    LANServerAvailable                          = 4511

    ReportWarKill                               = 5001
    VoteKickBanPlayer                           = 5018
    VoteKickBanPlayerResult                     = 5019
    FreeTrialChooseMostHelpfulFriend            = 5022
    RequestTF2Friends                           = 5023
    RequestTF2FriendsResponse                   = 5024
    ReplaySubmitContestEntry                    = 5026
    ReplaySubmitContestEntryResponse            = 5027
    SaxxyAwarded                                = 5029
    FreeTrialThankedBySomeone                   = 5028
    FreeTrialThankedSomeone                     = 5030
    FreeTrialConvertedToPremium                 = 5031
    CoachingAddToCoaches                        = 5200
    CoachingAddToCoachesResponse                = 5201
    CoachingRemoveFromCoaches                   = 5202
    CoachingRemoveFromCoachesResponse           = 5203
    CoachingFindCoach                           = 5204
    CoachingFindCoachResponse                   = 5205
    CoachingAskCoach                            = 5206
    CoachingAskCoachResponse                    = 5207
    CoachingCoachJoinGame                       = 5208
    CoachingCoachJoining                        = 5209
    CoachingCoachJoined                         = 5210
    CoachingLikeCurrentCoach                    = 5211
    CoachingRemoveCurrentCoach                  = 5212
    CoachingAlreadyRatedCoach                   = 5213
    DuelRequest                                 = 5500
    DuelResponse                                = 5501
    DuelResults                                 = 5502
    DuelStatus                                  = 5503
    HalloweenReservedItem                       = 5607
    HalloweenGrantItem                          = 5608
    HalloweenGrantItemResponse                  = 5609
    HalloweenServerBossEvent                    = 5612
    HalloweenMerasmus2012                       = 5613
    HalloweenUpdateMerasmusLootLevel            = 5614
    GameServerLevelInfo                         = 5700
    GameServerAuthChallenge                     = 5701
    GameServerAuthChallengeResponse             = 5702
    GameServerCreateIdentity                    = 5703
    GameServerCreateIdentityResponse            = 5704
    GameServerList                              = 5705
    GameServerListResponse                      = 5706
    GameServerAuthResult                        = 5707
    GameServerResetIdentity                     = 5708
    GameServerResetIdentityResponse             = 5709
    ClientUseServerModificationItem             = 5710
    ClientUseServerModificationItemResponse     = 5711
    GameServerUseServerModificationItem         = 5712
    GameServerUseServerModificationItemResponse = 5713
    GameServerServerModificationItemExpired     = 5714
    GameServerModificationItemState             = 5715
    GameServerAckPolicy                         = 5716
    GameServerAckPolicyResponse                 = 5717
    QPScoreServers                              = 5800
    QPScoreServersResponse                      = 5801
    QPPlayerJoining                             = 5802
    GameMatchSignOut                            = 6204
    CreateOrUpdateParty                         = 6233
    AbandonCurrentGame                          = 6235
    ForceSOCacheResend                          = 6237
    RequestChatChannelList                      = 6260
    RequestChatChannelListResponse              = 6261
    ReadyUp                                     = 6270
    KickedFromMatchmakingQueue                  = 6271
    LeaverDetected                              = 6272
    LeaverDetectedResponse                      = 6287
    PlayerFailedToConnect                       = 6288
    ExitMatchmaking                             = 6289
    AcceptInvite                                = 6291
    AcceptInviteResponse                        = 6292
    MatchmakingProgress                         = 6293
    MvMVictoryInfo                              = 6294
    GameServerMatchmakingStatus                 = 6295
    CreateOrUpdatePartyReply                    = 6296
    MvMVictory                                  = 6297
    MvMVictoryReply                             = 6298
    GameServerKickingLobby                      = 6299
    LeaveGameAndPrepareToJoinParty              = 6300
    RemovePlayerFromLobby                       = 6301
    SetLobbySafeToLeave                         = 6302
    UpdatePeriodicEvent                         = 6400
    ClientVerificationChallenge                 = 6500
    ClientVerificationChallengeResponse         = 6501
    ClientVerificationVerboseResponse           = 6502
    ClientSetItemSlotAttribute                  = 6503
    PlayerSkillRatingAdjustment                 = 6504
    WarIndividualUpdate                         = 6505
    WarJoinWar                                  = 6506
    WarRequestGlobalStats                       = 6507
    WarGlobalStatsResponse                      = 6508
    WorldStatusBroadcast                        = 6518
    DevGrantWarKill                             = 10001



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Any

from ._const import ReadOnly, impl_eq_via_id
from .abc import Awardable, Commentable, _CommentableKwargs
from .app import PartialApp
from .enums import Language, ReviewType
from .reaction import AwardReaction
from .user import ClientUser, User, WrapsUser
from .utils import DateTime

if TYPE_CHECKING:
    from typing_extensions import Self

    from .protobufs.reviews import RecommendationDetails as ReviewProto
    from .state import ConnectionState

__all__ = (
    "Review",
    "ReviewUser",
)


class ReviewUser(WrapsUser):
    __slots__ = (
        "number_of_apps_owned",
        "number_of_reviews",
        "playtime_forever",
        "playtime_last_two_weeks",
        "playtime_at_review",
        "last_played",
    )

    number_of_apps_owned: int | None
    number_of_reviews: int | None
    playtime_forever: timedelta
    playtime_last_two_weeks: timedelta
    playtime_at_review: timedelta
    last_played: datetime

    @classmethod
    def _from_proto(cls, state: ConnectionState, user: User | ClientUser, review: ReviewProto) -> Self:
        review_user = cls(state, user)
        review_user.number_of_reviews = None
        review_user.number_of_apps_owned = None
        review_user.playtime_forever = timedelta(seconds=review.playtime_forever)
        review_user.playtime_last_two_weeks = timedelta(seconds=review.playtime_2weeks)
        review_user.playtime_at_review = timedelta(seconds=review.playtime_at_review)
        review_user.last_played = DateTime.from_timestamp(review.last_playtime)
        return review_user

    @classmethod
    def _from_data(cls, state: ConnectionState, user: User | ClientUser, data: dict[str, Any]) -> Self:
        review_user = cls(state, user)
        review_user.number_of_reviews = data["num_reviews"]
        review_user.number_of_apps_owned = data["num_games_owned"]
        review_user.playtime_forever = timedelta(seconds=data["playtime_forever"])
        review_user.playtime_last_two_weeks = timedelta(seconds=data["playtime_last_two_weeks"])
        review_user.playtime_at_review = timedelta(seconds=data["playtime_at_review"])
        review_user.last_played = DateTime.from_timestamp(data["last_played"])
        return review_user


class ReviewApp(PartialApp):
    __slots__ = ("review_status",)

    def __init__(self, state: ConnectionState, id: int, review_status: int):
        super().__init__(state, id=id)
        self.review_status = ReviewType.try_value(review_status)


@dataclass(repr=False, eq=False)
@impl_eq_via_id
class Review(Commentable, Awardable):
    """Represents a review for an app."""

    __slots__ = (
        "_state",
        "id",
        "author",
        "app",
        "language",
        "content",
        "created_at",
        "updated_at",
        "recommended",
        "votes_helpful",
        "votes_funny",
        "weighted_vote_score",
        "commentable",
        "comment_count",
        "steam_purchase",
        "received_compensation",
        "written_during_early_access",
        "developer_response",
        "developer_responded_at",
        "reactions",
    )

    _state: ConnectionState
    id: ReadOnly[int]
    """The ID of the review."""
    author: ReviewUser
    """The author of the review."""
    app: ReviewApp
    """The app being reviewed."""
    recommended: bool
    """Whether the reviewer recommends the app."""
    language: Language
    """The language the review is written in."""
    content: str
    """The contents of the review."""
    created_at: datetime
    """The time the review was created at."""
    updated_at: datetime
    """The time the review was last updated at."""
    votes_helpful: int
    """The amount of people that marked the review as helpful."""
    votes_funny: int
    """The amount of people that marked this review as funny."""
    weighted_vote_score: float
    """The weighted score from Steam."""
    commentable: bool | None
    """Whether the author has allowed comments on the review."""
    comment_count: int
    """The amount of comments this review has."""
    steam_purchase: bool | None
    """Whether the app was purchased through Steam."""
    received_compensation: bool
    """Whether the reviewer received compensation for this review."""
    written_during_early_access: bool
    """Whether the reviewer played the app the app during early access."""
    developer_response: str | None
    """The developer's response to the review."""
    developer_responded_at: datetime | None
    """The time the developer responded to the review."""
    reactions: list[AwardReaction] | None
    """The review's reactions."""

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} id={self.id} app={self.app!r} author={self.author!r}>"

    @property
    def _commentable_kwargs(self) -> _CommentableKwargs:
        return {
            "id64": self.author.id64,
            "forum_id": self.app.id,
        }

    @classmethod
    def _from_proto(cls, state: ConnectionState, review: ReviewProto, user: User | ClientUser) -> Self:
        return cls(
            state,
            id=review.recommendationid,
            author=ReviewUser._from_proto(state, user, review),
            app=ReviewApp(state, review.appid, review.review_quality),
            language=Language.from_str(review.language),
            content=review.review,
            created_at=DateTime.from_timestamp(review.time_created),
            updated_at=DateTime.from_timestamp(review.time_updated),
            recommended=review.voted_up,
            votes_funny=review.votes_funny,
            votes_helpful=review.votes_up,
            weighted_vote_score=review.weighted_vote_score,
            commentable=not review.comments_disabled,
            comment_count=review.comment_count,
            steam_purchase=not review.unverified_purchase,
            received_compensation=review.received_compensation,
            written_during_early_access=review.written_during_early_access,
            developer_response=review.developer_response or None,
            developer_responded_at=(
                DateTime.from_timestamp(review.time_developer_responded)
                if review.time_developer_responded
                else None  # TODO add the rest of the developer attrs
            ),
            reactions=[AwardReaction(state, reaction) for reaction in review.reactions],
        )

    @classmethod
    def _from_data(cls, state: ConnectionState, data: dict[str, Any], app: ReviewApp, user: User) -> Self:
        return cls(
            state,
            id=data["recommendationid"],
            author=ReviewUser._from_data(state, user, data["author"]),
            app=app,
            language=Language.from_str(data["language"]),
            content=data["review"],
            created_at=DateTime.from_timestamp(data["timestamp_created"]),
            updated_at=DateTime.from_timestamp(data["timestamp_updated"]),
            recommended=data["voted_up"],
            votes_funny=data["votes_funny"],
            votes_helpful=data["votes_up"],
            weighted_vote_score=data["weighted_vote_score"],
            commentable=None,
            comment_count=data["comment_count"],
            steam_purchase=data["steam_purchase"],
            received_compensation=data["received_for_free"],
            written_during_early_access=data["written_during_early_access"],
            developer_response=data.get("developer_response"),
            developer_responded_at=(
                DateTime.from_timestamp(data["timestamp_dev_responded"]) if "timestamp_dev_responded" in data else None
            ),
            reactions=None,
        )

    async def mark_helpful(self) -> None:
        """Mark this review as helpful."""
        await self._state.http.mark_review_as_helpful(self.id, True)

    async def mark_not_helpful(self) -> None:
        """Mark this review as not helpful."""
        await self._state.http.mark_review_as_helpful(self.id, False)

    async def mark_funny(self) -> None:
        """Mark this review as funny."""
        await self._state.http.mark_review_as_funny(self.id)

    async def edit(
        self,
        content: str,
        /,
        *,
        public: bool = True,
        commentable: bool | None = None,
        language: Language | None = None,
        written_during_early_access: bool | None = None,
        received_compensation: bool | None = None,
    ) -> None:
        """Edit this review.

        Parameters
        ----------
        content
            The content of the review.
        public
            Whether the review should be shown publicly on your profile.
        commentable
            Whether the review should be commentable.
        language
            The language of the review. Defaults to the current value.
        written_during_early_access
            Whether the review was written during early access. Defaults to the current value.
        received_compensation
            Whether the user received compensation for this review. Defaults to the current value.
        """
        if self.commentable is None and commentable is None:
            review = await self.author.fetch_review(self.app)
            commentable = review.commentable
        await self._state.edit_review(
            self.id,
            content,
            public,
            self.commentable if commentable is None else commentable,  # type: ignore
            (self.language if language is None else language).api_name,  # TODO check
            self.written_during_early_access if written_during_early_access is None else written_during_early_access,
            self.received_compensation if received_compensation is None else received_compensation,
        )

    async def delete(self) -> None:
        """Delete this review."""
        await self._state.http.delete_review(self.app.id)
